/***************************************************************************
 *
 * Copyright (c) 2002 Promise Technology, Inc.  All rights reserved.
 * No part of this document may be reproduced or transmitted in any form or
 * by any means, electronic or mechanical, for any purpose, without the
 * express written permission of Promise Technology, Inc.
 *
 *
 * cam.c - main functions
 *
 ***************************************************************************/
#include "cam_con.h"
#include "cam_def.h"
#include "cam_imp.h"
#include "cam_macr.h"
#include "cam_gb.h"
#include "cam_exp.h"

#ifdef	_MMIO_

void WriteIndexUchar(PTR_SIZE BaseAddress, U8 Index, U8 Data )
{
	camWriteRegByte(BaseAddress+Index, Data);
}

U8 ReadIndexUchar(PTR_SIZE BaseAddress, U8 Index)
{
	return(camReadRegByte(BaseAddress+Index));
}

#else

void WriteIndexUchar(PTR_SIZE BaseAddress, U8 Index, U8 Data )
{
	camOutPortByte(BaseAddress+ INDEX_REGISTER_OFFSET, Index);
	camOutPortByte(BaseAddress+ DATA_REGISTER_OFFSET+(Index&0x03), Data);
}

U8 ReadIndexUchar(PTR_SIZE BaseAddress, U8 Index )
{
	camOutPortByte(BaseAddress+ INDEX_REGISTER_OFFSET, Index);
	return( camInPortByte(BaseAddress+ DATA_REGISTER_OFFSET+(Index&0x03)));
}

#endif

void CAMInitSEQIDPool(U8 Adapter_ID)
{
    U8 i;

    if (Adapter_ID > (MAX_ADAPTER_NUM -1))
        return;

    for (i = START_SEQ_ID; i < MAX_SEQ_ID; i++)
        SEQIDPool[Adapter_ID][i] = i;
        
    SEQIDCurrent[Adapter_ID] = START_SEQ_ID;

    return;
}


U8 CAMAllocSEQID(U8 Adapter_ID)
{
	U8 SEQID;

    if (Adapter_ID > (MAX_ADAPTER_NUM -1))
       	return U8NULL;

    if (SEQIDCurrent[Adapter_ID] >= MAX_SEQ_ID)
       	return U8NULL;

    SEQID = SEQIDPool[Adapter_ID][SEQIDCurrent[Adapter_ID]];
    SEQIDCurrent[Adapter_ID]++;

    return SEQID;
}

void CAMFreeSEQID(U8 Adapter_ID, U8 SEQID)
{

    if (Adapter_ID > (MAX_ADAPTER_NUM -1))
       	return;

    SEQIDCurrent[Adapter_ID]--;
    SEQIDPool[Adapter_ID][SEQIDCurrent[Adapter_ID]] = SEQID;

    return;
}

void CAMInitPacketQ(U8 Adapter_ID, PAC_Q_TYPE q_type)
{
	U8	 	i,j;
	PU8		base_addr;
	PTR_SIZE		basephyaddr;
	PADAPTER_CONFIG pCurAdapter = &(gAdapterConfig[Adapter_ID]);
	
	if (Adapter_ID > (MAX_ADAPTER_NUM -1))
        	return;
        
   	base_addr = (PU8)pCurAdapter->MemVirAddr;
	basephyaddr = pCurAdapter->MemPhyAddress;
	
	switch (q_type) {
		case ATA_PAC:
   			base_addr += ATA_BASE_ADDR;
	        basephyaddr = basephyaddr+ATA_BASE_ADDR+PACKET_HEAD_SIZE;
	        for (i = (Adapter_ID*MAX_ATA_CHNL); i < ((Adapter_ID+1)*MAX_ATA_CHNL); i++) {
    	    	for (j = 0; j < ATAPacketCount[Adapter_ID]; j++) {
					ATAPacketQueue[i][j] = (PCAM_PACKET)base_addr;
					(ATAPacketQueue[i][j])->PacketHead.Address = basephyaddr;
					base_addr += sizeof(CAM_PACKET);
					basephyaddr += ATA_PAC_SIZE;
				}
        		ATAPacketCurrent[i] = 0;
        		ATAPacketHead[i] = 0;
        		ATAPacketTail[i] = 0;
        		//ATAPacketTail[i] = ATAPacketCount[Adapter_ID] - 1;
        	}
	        break;
	    default:
    	    return;
    }
    return;
}

PVOID CAMAllocCamPacket(U8 bChannelNumber)
{
    PVOID	PacketAddress;
    U8 		Adapter_ID=bChannelNumber/MAX_ATA_CHNL;

	if (bChannelNumber >= MAX_ATA_MODULE)
		return ((PVOID)PTR_NULL);

 	if (ATAPacketCurrent[bChannelNumber] >= ATAPacketCount[Adapter_ID])
  		return ((PVOID)PTR_NULL);


   	PacketAddress = (PVOID)ATAPacketQueue[bChannelNumber][ATAPacketHead[bChannelNumber]];
   	ATAPacketCurrent[bChannelNumber]++;
   	ATAPacketHead[bChannelNumber]++;
   	ATAPacketHead[bChannelNumber] %= ATAPacketCount[Adapter_ID];

	((PCAM_PACKET)PacketAddress)->PacketHead.Flag = ZERO_NULL;

    return PacketAddress;
}

void CAMFreeCamPacket(U8 bChannelNumber , PVOID PacketAddress)
{
	U8 		Adapter_ID=bChannelNumber/MAX_ATA_CHNL;
	
	if (bChannelNumber >= MAX_ATA_MODULE)
		return;

	if(ATAPacketCurrent[bChannelNumber] == 0)
		return;
	
	ATAPacketCurrent[bChannelNumber]--;
	ATAPacketQueue[bChannelNumber][ATAPacketTail[bChannelNumber]] = (PCAM_PACKET)PacketAddress;

   	ATAPacketTail[bChannelNumber]++;
   	ATAPacketTail[bChannelNumber] %= ATAPacketCount[Adapter_ID];

	return;
}


void CAMInitATAPacketQ(U8 Adapter_ID, PAC_Q_TYPE q_type)
{
	U8	 	i,j;
	
	if (Adapter_ID > (MAX_ADAPTER_NUM -1))
		return;

	switch (q_type) {
		case ATA_PAC:
			for (i = (Adapter_ID*MAX_ATA_CHNL); i < ((Adapter_ID+1)*MAX_ATA_CHNL); i++) {
				for (j = 0; j < MAX_ATA_RECEIVE_PAC; j++) {
					//ATATaskPacketQ[i][j] = (PCAM_ATA_PACKET)(&ATATaskPacketQPool[i][j]);
					ATATaskPacketQ[i][j] = j;
				}
				ReceATAPacketCurrent[i] = 0;
			}
			break;
		default:
			return;
	}
	return;
}


PVOID CAMAllocATAPacket(U8 bChannelNumber)
{
	PVOID	PacketAddress;
	U8		PoolIndex;		
	/*U8 		Adapter_ID=bChannelNumber/MAX_ATA_CHNL;*/

	if (bChannelNumber >= MAX_ATA_MODULE)
		return ((PVOID)PTR_NULL);

 	if (ReceATAPacketCurrent[bChannelNumber] >= MAX_ATA_RECEIVE_PAC)
  		return ((PVOID)PTR_NULL);

	PoolIndex = ATATaskPacketQ[bChannelNumber][ReceATAPacketCurrent[bChannelNumber]];
   	PacketAddress = (PVOID)(&ATATaskPacketQPool[bChannelNumber][PoolIndex]);
   	
   	//PacketAddress = (PVOID)ATATaskPacketQ[bChannelNumber][ReceATAPacketCurrent[bChannelNumber]];
   	ReceATAPacketCurrent[bChannelNumber]++;

	((PCAM_ATA_PACKET)PacketAddress)->PacketHead.Flag = ZERO_NULL;

	return PacketAddress;
}

void CAMFreeATAPacket(U8 bChannelNumber , PVOID PacketAddress)
{
	PTR_SIZE	PoolAddr=(PTR_SIZE)PacketAddress;
	
	if (bChannelNumber >= MAX_ATA_MODULE)
		return;

 	if (ReceATAPacketCurrent[bChannelNumber] <= 0)
 		return;

   	ReceATAPacketCurrent[bChannelNumber]--;
   	//ATATaskPacketQ[bChannelNumber][ReceATAPacketCurrent[bChannelNumber]] = (PCAM_ATA_PACKET)PacketAddress;

	PoolAddr -= (PTR_SIZE)(ATATaskPacketQPool[bChannelNumber]);
	PoolAddr /= RECE_PAC_SIZE;
	ATATaskPacketQ[bChannelNumber][ReceATAPacketCurrent[bChannelNumber]] = (U8)PoolAddr;
	return;
}

void CAMInitSGQ(U8 Adapter_ID, PAC_Q_TYPE q_type)
{
    U8	 	i,j;
	PU8		base_addr;
	PTR_SIZE		basephyaddr;
	PADAPTER_CONFIG pCurAdapter = &(gAdapterConfig[Adapter_ID]);
	
    if (Adapter_ID > (MAX_ADAPTER_NUM -1))
       	return;
        
   	base_addr = (PU8)pCurAdapter->MemVirAddr;
	basephyaddr = pCurAdapter->MemPhyAddress;


    switch (q_type) {
    	case ATA_PAC:
		base_addr = (PU8)((PTR_SIZE)base_addr+ATA_BASE_ADDR+((PTR_SIZE)ATAPacketCount[Adapter_ID]*MAX_ATA_CHNL * ATA_PAC_SIZE));
		basephyaddr = basephyaddr+ATA_BASE_ADDR+((PTR_SIZE)ATAPacketCount[Adapter_ID]*MAX_ATA_CHNL * ATA_PAC_SIZE);    	
        	for (i = (Adapter_ID*MAX_ATA_CHNL); i < ((Adapter_ID+1)*MAX_ATA_CHNL); i++) {
        		for (j = 0; j < ATAPacketCount[Adapter_ID]; j++) {
           			PCAM_ATA_SG_Queue[i][j] = (PCAM_ATA_SG)base_addr;
            		(PCAM_ATA_SG_Queue[i][j])->Address = basephyaddr;
           			//((PCAM_ATA_SG)base_addr)++;
           			base_addr += sizeof(CAM_ATA_SG);
           			basephyaddr += ATA_SG_SIZE;
        		}
        		CAM_ATA_SG_Current[i] = 0;
        		CAM_ATA_SG_Head[i] = 0;        		
        		CAM_ATA_SG_Tail[i] = 0;
        	}
       		break;
    	default:
       		return;
   	}
   	return;
}

PVOID CAMAllocATASG(U8 bChannelNumber)
{
   	PVOID	PacketAddress;
   	U8 		Adapter_ID=bChannelNumber/MAX_ATA_CHNL;

	if (bChannelNumber >= MAX_ATA_MODULE)
		return ((PVOID)PTR_NULL);
		
   	if (CAM_ATA_SG_Current[bChannelNumber] >= ATAPacketCount[Adapter_ID])
 		return ((PVOID)PTR_NULL);

   	PacketAddress = (PVOID) PCAM_ATA_SG_Queue[bChannelNumber][CAM_ATA_SG_Head[bChannelNumber]];
   	CAM_ATA_SG_Current[bChannelNumber]++;
   	CAM_ATA_SG_Head[bChannelNumber]++;
	CAM_ATA_SG_Head[bChannelNumber] %= ATAPacketCount[Adapter_ID];
	
   	return PacketAddress;
}


void CAMFreeATASG(U8 bChannelNumber, PVOID PacketAddress)
{
	U8 		Adapter_ID=bChannelNumber/MAX_ATA_CHNL;
	
	if (bChannelNumber >= MAX_ATA_MODULE)
       	return;

   	if (CAM_ATA_SG_Current[bChannelNumber] == 0)
 		return;

	CAM_ATA_SG_Current[bChannelNumber]--;
	PCAM_ATA_SG_Queue[bChannelNumber][CAM_ATA_SG_Tail[bChannelNumber]] = (PCAM_ATA_SG)PacketAddress;

	CAM_ATA_SG_Tail[bChannelNumber]++;
	CAM_ATA_SG_Tail[bChannelNumber] %= ATAPacketCount[Adapter_ID];

   	return;
}

/* if p_S_SG->Reserved !=0, set to zero, and if p_S_SG->CtrlFlag & 80 != 0, clear the bits. */
/* copy source p_S_SG to target p_T_SG (contiguous physical memory). */
void checkConvertSG(PCAMSG  p_S_SG,PCAMSG p_T_SG)
{
	while(1){
		p_T_SG->Reserved=0;
		
		/* for Big-Endian */
		p_T_SG->wCount = CamByteSwap16(p_S_SG->wCount);
		p_T_SG->StartAddr = CamByteSwap32(p_S_SG->StartAddr);

		if(p_S_SG->CtrlFlag & EOT){
			p_T_SG->CtrlFlag = EOT;
			break;		
		}
		else{
			p_T_SG->CtrlFlag = 0;		
		}
				
		p_S_SG++;
		p_T_SG++;
	}
			
	return;
}

void CAMInitSerialBusAddr(U8 Adapter_ID)
{
	PU8		base_addr;
	PTR_SIZE		basephyaddr;
	PADAPTER_CONFIG pCurAdapter = &(gAdapterConfig[Adapter_ID]);
	
    if (Adapter_ID > (MAX_ADAPTER_NUM -1))
       	return;
        
   	base_addr = (PU8)pCurAdapter->MemVirAddr;
	basephyaddr = pCurAdapter->MemPhyAddress;
	
	SerialBusAddr[Adapter_ID].MemPhyAddress[0]=(U32)(basephyaddr+SREIAL_BUS_BUFFER_ADDR);
	SerialBusAddr[Adapter_ID].MemVirAddr[0]=(PVOID)(base_addr+SREIAL_BUS_BUFFER_ADDR);

	SerialBusAddr[Adapter_ID].MemPollPhyAddress[0]=(U32)(basephyaddr+SREIAL_BUS_BUFFER_ADDR+SREIAL_BUS_BUFFER_SIZE);
	SerialBusAddr[Adapter_ID].MemPollPhyAddress[1]=(U32)(basephyaddr+SREIAL_BUS_BUFFER_ADDR+SREIAL_BUS_BUFFER_SIZE+SREIAL_BUS_POLLING_BUFFER_SIZE);
	SerialBusAddr[Adapter_ID].MemPollVirAddr[0]=(PVOID)(base_addr+SREIAL_BUS_BUFFER_ADDR+SREIAL_BUS_BUFFER_SIZE);
	SerialBusAddr[Adapter_ID].MemPollVirAddr[1]=(PVOID)(base_addr+SREIAL_BUS_BUFFER_ADDR+SREIAL_BUS_BUFFER_SIZE+SREIAL_BUS_POLLING_BUFFER_SIZE);
		
	SerialBusAddr[Adapter_ID].MemVirCurrent=SerialBusAddr[Adapter_ID].MemVirAddr[0];
	SerialBusCurrent[Adapter_ID]=0;
	SerialBusCMD=FALSE;
   	return;
}

void Put2SeqIDQueue(U8 Adapter_ID,PVOID PacketAddress,U8 SeqID,U8 Set_ulNCA)
{

	PSEQ_INFO	pCurSEQ_INFO;
	PCAM_PACKET ultmpATAPacketAddress;
	
	((PCAM_PACKET)PacketAddress)->PacketHead.NPA=(PVOID)PTR_NULL;
	
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SeqID]);


#ifdef _ERR_INJECT_
{
	PATA_TASK	pATATask;

	ultmpATAPacketAddress=(PCAM_PACKET)PacketAddress;
	pATATask = ultmpATAPacketAddress->PacketHead.PTA;

	if(ErrInj_Buffer_Table[pATATask->bCh2Chip].count){	
		if(ultmpATAPacketAddress->PacketHead.Flag & PACFLAG_ERR_INJECT){
			if(!(ErrInject_SEQ[Adapter_ID] & (1 << SeqID))){
				ErrInject_SEQ[Adapter_ID] |= (1 << SeqID);		
				if(ultmpATAPacketAddress->PacketHead.Flag & PACFLAG_ERR_INJECT)
					ErrInject_SEQ_Timeout[Adapter_ID] |= (1 << SeqID);
			}
		}
	}
}
#endif
	


	
	if(pCurSEQ_INFO->SEQID2PacketHead==(PVOID)PTR_NULL){
		PATA_TASK pATATask;
		ultmpATAPacketAddress=(PCAM_PACKET)PacketAddress;
		pATATask = (PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
		pCurSEQ_INFO->Owner=(ModType) (pATATask->bCh2Chip%MAX_ATA_CHNL);
		
		pCurSEQ_INFO->PACCNT = 1;
		pCurSEQ_INFO->SEQID2PacketHead=PacketAddress;
		pCurSEQ_INFO->SEQID2PacketTail=PacketAddress;
	}
	else{
		
		ultmpATAPacketAddress=(PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketTail;
		
		pCurSEQ_INFO->SEQID2PacketTail=PacketAddress;
		ultmpATAPacketAddress->PacketHead.NPA=PacketAddress;
		if (Set_ulNCA==TRUE) {
			/* pointer to next ATA Packet */
			ultmpATAPacketAddress->ATAPacket.ulNCA = CamByteSwap32((U32)((PCAM_PACKET)PacketAddress)->PacketHead.Address);
			pCurSEQ_INFO->PACCNT++;
		}
		
	}			
}


#ifdef Encryption
U8	CAM_EncryptionKeySearch(U8 Adapter_ID)
{
	U8		j=0, returnStatus=FALSE;
	U8		EProm_data[8];
	U8		CheckDataSum;
	U32		dwData[2];


//	WritePortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, 0x0F0F);
//	
//	/* prepare SBUS buffer */
//	Buffer=(PU8)(SerialBusAddr[Adapter_ID].MemPollVirAddr[0]);
//	j=0;
//	*(Buffer+j)=0xA0; /* bit 0 : R/W 1/0 */ /* EP-Rom data Address  0xA0 */
//	j++;
//	*(Buffer+j)=0x80; /* Control */
//	j++;
//	*(Buffer+j)=5; /* Byte Count Low */
//	j++;	
//	*(Buffer+j)=0; /* Byte Count High */
//	j++;
//	*(Buffer+j)=0;
//	j++;
//	*(Buffer+j)=0x10;
//	j++;
//	*(Buffer+j)=0x5A;
//	j++;
//	*(Buffer+j)=0x11;
//	j++;
//	*(Buffer+j)=0x22;
//
//
//	bData = Data2SerialBus(Adapter_ID, (U32)SerialBusAddr[Adapter_ID].MemPollPhyAddress[0]);
//	if (bData != TRUE)	/* ERROR done */
//	{
//		printf("ERROR 1\n");
//		return returnStatus;		
//	}
//
//
//
//
//	/* prepare SBUS buffer */
//	Buffer=(PU8)(SerialBusAddr[Adapter_ID].MemPollVirAddr[0]);
//	j=0;
//	*(Buffer+j)=0xA0; /* bit 0 : R/W 1/0 */ /* EP-Rom data Address  0xA0 */
//	j++;
//	*(Buffer+j)=0x80; /* Control */
//	j++;
//	*(Buffer+j)=5; /* Byte Count Low */
//	j++;	
//	*(Buffer+j)=0; /* Byte Count High */
//	j++;
//	*(Buffer+j)=4;
//	j++;
//	*(Buffer+j)=0x33;
//	j++;
//	*(Buffer+j)=0x44;
//	j++;
//	*(Buffer+j)=0xA5;
//	j++;
//	*(Buffer+j)=0x47;
//
//
//	*((PU32)(Buffer+j)) = 0x22115a10;
//	j+=4;
//	*((PU32)(Buffer+j)) = 0x47A54433;
//
//	bData = Data2SerialBus(Adapter_ID, (U32)SerialBusAddr[Adapter_ID].MemPollPhyAddress[0]);
//	if (bData != TRUE)	/* ERROR done */
//	{
//		printf("ERROR 2\n");
//		return returnStatus;		
//	}


	
	CAM_SerialBusWriteDATAPolling(Adapter_ID,0x00A0,2);
	dwData[0]=CAM_SerialBusReadDATAPolling(Adapter_ID,0xA0,4);

	EProm_data[0] = (U8)dwData[0];	
	EProm_data[1] = (U8)(dwData[0]>>8);	
	EProm_data[2] = (U8)(dwData[0]>>16);	
	EProm_data[3] = (U8)(dwData[0]>>24);	

	CAM_SerialBusWriteDATAPolling(Adapter_ID,0x04A0,2);
	dwData[1]=CAM_SerialBusReadDATAPolling(Adapter_ID,0xA0,4);
	
	EProm_data[4] = (U8)dwData[1];	
	EProm_data[5] = (U8)(dwData[1]>>8);	
	EProm_data[6] = (U8)(dwData[1]>>16);	
	EProm_data[7] = (U8)(dwData[1]>>24);	
	
	
	if((EProm_data[0] == 0x5A) && (EProm_data[1] == 0x10)){
		CheckDataSum = 0;
		for(j=0; j<8; j++)
			CheckDataSum += EProm_data[j];
		
		if(!CheckDataSum){
			gAdapterConfig[Adapter_ID].Features |= WITH_EncryKEY;
			gAdapterConfig[Adapter_ID].EncryptionKey = (dwData[0] >> 16) | (dwData[1] << 16);
			returnStatus = TRUE;
		}
	}
	return returnStatus;
}
#endif

/*---------------------------------------------------------------------------------*/
U8 CAM_ReInit(PADAPTER_CONFIG_INFO pAdapter)
{
	return(CAM_Init(pAdapter));
}

U8 CAM_Init(PADAPTER_CONFIG_INFO pAdapter)
{
	U8	WithParallel;
	U8	i, j;
	U8	bData;
	U8	CurAdapter_ID;
	U8	StartChnl;
	U8	EndChnl;
	U8	CurMaxATAChnl = MAX_ATA_CHNL;
	U32	lindex;
	PADAPTER_CONFIG pCurAdapter;
	PSEQ_INFO	pCurSEQ_INFO;
#ifdef	_MMIO_
	PTR_SIZE	base_addr;
#endif		
	PTR_SIZE 	AddressShift;
	U32	 	tempaddress;
	PTR_SIZE	tempAdr;

	if ( pAdapter->MemSize < MIN_CAM_MEMSIZE )
		return(camFAIL);

	if ( pAdapter->MemSize > MAX_CAM_MEMSIZE )
		pAdapter->MemSize = MAX_CAM_MEMSIZE;

#ifdef NCQ
	if(pAdapter->NoNCQAccess != TRUE){
		/* check physical memory is in 64K boundary */
		if ( (pAdapter->MemPhyAddress & 0xFFFF0000) != ((pAdapter->MemPhyAddress+pAdapter->MemSize) & 0xFFFF0000) )
			return(camFAIL);
	}

#endif
		
	/* save adapter data */
	CurAdapter_ID = pAdapter->Adapter_ID;
	StartChnl = CurAdapter_ID*MAX_ATA_CHNL;
	EndChnl = StartChnl + MAX_ATA_CHNL - 1;
	pCurAdapter = &(gAdapterConfig[CurAdapter_ID]);
	for ( lindex = 0; lindex < sizeof(ADAPTER_CONFIG); lindex++ )
	{
		*(((PU8)pCurAdapter)+lindex) = 0;
	}
	pCurAdapter->ATABaseAddr = pAdapter->BaseAddress[0];
	pCurAdapter->XORBaseAddr = pAdapter->BaseAddress[1];
	pCurAdapter->HostBaseAddr = pAdapter->BaseAddress[2];
	pCurAdapter->MemIOBaseAddr = pAdapter->BaseAddress[3];
	pCurAdapter->MemPhyAddress = pAdapter->MemPhyAddress; 
	pCurAdapter->MemVirAddr = pAdapter->MemVirAddr; 
	pCurAdapter->MemSize = pAdapter->MemSize-4; 

	pCurAdapter->TapeOutVersion = pAdapter->TapeOutVersion; 
	pCurAdapter->NoNCQAccess = pAdapter->NoNCQAccess; 
	pCurAdapter->Features = 0;
	pCurAdapter->EncryptionKey = U32NULL;

	
	tempAdr = (PTR_SIZE)pCurAdapter->MemVirAddr;
#ifdef	_64BIT_
	/* set QWord boundary */
	if((AddressShift=(tempAdr&0x7)) != 0){
		pCurAdapter->MemVirAddr = (PVOID)(tempAdr + (8-AddressShift));
		pCurAdapter->MemPhyAddress = pAdapter->MemPhyAddress+(8-AddressShift); 
	}
#else
	/* set DWord boundary */
	if((AddressShift=(tempAdr&0x3)) != 0){
          		pCurAdapter->MemVirAddr = (PVOID)(tempAdr + (4-AddressShift));
		pCurAdapter->MemPhyAddress = pAdapter->MemPhyAddress+(4-AddressShift); 
	}      
#endif
	
	for ( lindex = 0; lindex < pCurAdapter->MemSize; lindex++ )
		*((PU8)pCurAdapter->MemVirAddr+lindex) = 0;	/* clear memory*/
	
	/* save module data */
#ifdef	_MMIO_
	base_addr = pCurAdapter->MemIOBaseAddr;
   	base_addr += ATAMEMBASE;
	for (i = StartChnl,j=0; i <= EndChnl; i++, j++ ) {
		ATABaseAddr[i] = base_addr + ((U32)j << 7); 			/* 200/280/300/380 */
		ATA_ExpBaseAddr[i]= base_addr + (j * 0x100)+ 0x280; 	/* 480/580/680/780 */
	}
	HostBaseAddr[CurAdapter_ID] = pCurAdapter->MemIOBaseAddr + HOSTMEMBASE;		
#else
	for (i = StartChnl, j = 0; i <= EndChnl; i++, j++ ) {
		ATABaseAddr[i] = pCurAdapter->ATABaseAddr + ((U32)j << 4);
		ATA_ExpBaseAddr[i] = pCurAdapter->ATABaseAddr + ((U32)j << 3) + 0x38;
	}
	HostBaseAddr[CurAdapter_ID] = pCurAdapter->HostBaseAddr;
#endif

	/* calculate ATAPacketCount */
	ATAPacketCount[CurAdapter_ID]=(U8)((pCurAdapter->MemSize-MIN_CAM_MEMSIZE+4)/CAM_PACKET_SIZE)+1;
	
#ifdef	_MMIO_
	/* for flash memory interface */
	gFlashConfig[CurAdapter_ID].FlashMemAddr = pAdapter->BaseAddress[4];
	gFlashConfig[CurAdapter_ID].Flag = FLASH_320;
#endif		

	/* eanble BMR_BURST */
	ReadPortUlong( HostBaseAddr[CurAdapter_ID], oFFSETHOSTFlashCTL, tempaddress );
	tempaddress |= 0x00002000;
	WritePortUlong(HostBaseAddr[CurAdapter_ID], oFFSETHOSTFlashCTL, tempaddress );
		
	/* clear plug/unplug flags */
	ReadPortUlong( HostBaseAddr[CurAdapter_ID], oFFSETHOSTSATAPLUG, tempaddress );	
	tempaddress |= 0xFF;		/* clear plug/unplug flags */
	WritePortUlong( HostBaseAddr[CurAdapter_ID], oFFSETHOSTSATAPLUG, tempaddress );	
	
	/* unmask plug/unplug ints */
	ReadPortUlong( HostBaseAddr[CurAdapter_ID], oFFSETHOSTSATAPLUG, tempaddress );	
	tempaddress &= 0xFF00FFFF;		/* clear plug/unplug flags */
	WritePortUlong( HostBaseAddr[CurAdapter_ID], oFFSETHOSTSATAPLUG, tempaddress );

	WithParallel = FALSE;
	ReadPortUchar(HostBaseAddr[CurAdapter_ID], oFFSETHOSTFlashCTL+1, bData );
	if (!(bData & 0x80))
	{
#if defined(OEM_OMB_TUNE_EYE)
		/* for onboard OEM required, Ch0:0x540; Ch1:0x640 */
		U32  templdata1;
		for (i = StartChnl; i<= StartChnl+2;i++ ){
	#ifdef	_MMIO_
			ReadIndexUlong(ATABaseAddr[(i/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (i%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, templdata1);
			if(i == StartChnl)
				templdata1 |= 0x00000100; 
			else
				templdata1 |= 0x00000200; 
			WriteIndexUlong(ATABaseAddr[(i/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (i%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, templdata1);
	#else
			ReadIndexUlong(ATABaseAddr[i], iNDEXIDEPHYMode2Reg, templdata1);
			if(i == StartChnl)
				templdata1 |= 0x00000100; 
			else
				templdata1 |= 0x00000200; 
			WriteIndexUlong(ATABaseAddr[i], iNDEXIDEPHYMode2Reg, templdata1);			
	#endif
		}
#endif
		/* PDC20575 series ASIC */
		WithParallel = TRUE;
	}
	
	/* initialize ModuleStatus,QHeads,QTails,QCounts */
	
	CAMInitSEQIDPool(CurAdapter_ID);
	
	/* initialize the ata queue */
	for( i = StartChnl; i <= EndChnl; i++ ){
		ATAQueueHead[i] = (PCAM_ATA_PACKET)PTR_NULL;
		ATAQueueTail[i] = (PCAM_ATA_PACKET)PTR_NULL;
		for( j=0;j<MAX_ID_PER_CHNL;j++ ){
			ATASortHead[i][j] = (PCAM_ATA_PACKET)PTR_NULL;
			ATASortLBAl[i][j] = ZERO_NULL;
			ATASortLBAh[i][j] = ZERO_NULL;			
		}
		ATAModuleBusy[i] = FALSE;
		ATAModuleNeedPush[i] = FALSE;
		ATAModuleSEQID[i] = U8NULL;
		NeedCheckFIFO[i] = FALSE;
	}	
	/* initialization for sequence control */
	pCurSEQ_INFO = &(gSEQ_INFO[CurAdapter_ID][0]);
	for( i = 0; i < MAX_SEQ_ID; i++ )
	{
		pCurSEQ_INFO->SEQID2PacketHead = (PVOID)PTR_NULL;
		pCurSEQ_INFO->SEQID2PacketTail = (PVOID)PTR_NULL;
		pCurSEQ_INFO++;
	}
	
	/* initialize packet Queue Count*/
	for (i=0; i< MAX_ATA_CHNL; i++)
		/* set default PacketQueueCount to 1 (not do packets queue). */
		ATAPacketQueueCount[CurAdapter_ID].ATAModule[i]=1;
		 /* ATAPacketQueueCount[CurAdapter_ID].ATAModule[i]=ATAPacketCount[CurAdapter_ID]; */

	/* initialize the BOX data */
	for( i = StartChnl; i <= EndChnl; i++ ){
		for ( lindex = 0; lindex < sizeof(ENCLOSURE_TYPE); lindex++ )
			*(((PU8)(&gBOXConfig[i]))+lindex) = 0;

		for ( lindex = 0; lindex < sizeof(ENCLOSURE_STATUS); lindex++ )
			*(((PU8)(&gBOXStatus[i]))+lindex) = 0;
	}
				
	/* initialize packet Queue */
	CAMInitPacketQ(CurAdapter_ID, ATA_PAC);
	
	/* initialize receive packet Queue */
	CAMInitATAPacketQ(CurAdapter_ID, ATA_PAC);

	/* initialize packet sg queue */
	CAMInitSGQ(CurAdapter_ID, ATA_PAC);

	/* initialize serial bus buffer */
	CAMInitSerialBusAddr(CurAdapter_ID);
		
	/* initialize ATA channel infomation */
	if ( WithParallel == TRUE ){
		if(pAdapter->Features & HBA_TYPE_2S)	
			CurMaxATAChnl = 2;	
		else	
			CurMaxATAChnl = 3;
	}
	
	for (i = StartChnl; i<= EndChnl;i++ )
	{	
		if ( i >= (StartChnl+CurMaxATAChnl) ) {
			gChnlConfig[i].bCh = camCHNL_NOT_FOUND;
			continue;
		}	
	
		gChnlConfig[i].bCh = i;
		
		if ( (i >= StartChnl+2 ) && ( WithParallel == TRUE ) ) {	
			gChnlConfig[i].ChnlType = CHNL_ATA;		/* parallel ATA channel */
			gChnlConfig[i].ChnlSpeed = CHNL_U133;	
			/* check cable type */
			bData = ReadIndexUchar( ATABaseAddr[i], (iNDEXIDECICR+3));
			if ( bData & 0x01 ){
				gChnlConfig[i].CableType = CABLE_40;
				gChnlConfig[i].ChnlSpeed = CHNL_U33;
			}
			else
				gChnlConfig[i].CableType = CABLE_80;
		}
		else {
			gChnlConfig[i].ChnlType = CHNL_SATA;
			gChnlConfig[i].ChnlSpeed = CHNL_U150;
			gChnlConfig[i].CableType = CABLE_SATA;
			gBOXConfig[i].BOXID = 0xFF;	/* no box */
			
			/* set suitable value for PHY quality */
			WriteIndexUlong(ATABaseAddr[i], iNDEXIDEPHYQuality, (U32)0x405500A0);			

		}
	}

	/* reset series bus state machine */
	CAM_ResetSerialBusMachine(CurAdapter_ID);
		
	/* initialize device info */	
	for( i = StartChnl; i<= EndChnl; i++ )
	{	
		for ( j =0; j < MAX_ID_PER_CHNL; j++ ) 
		{ 
			gDrvConfig[i][j].bCh = gChnlConfig[i].bCh;
			gDrvConfig[i][j].bID = j;
			gDrvConfig[i][j].DevFlag = DEV_NOTFOUND;
			gDrvConfig[i][j].pEnclosure = ZERO_NULL;
		}
		gChnlDrvControlConfig[i].TagCmdFlag = 0;
		gChnlDrvControlConfig[i].Current_bID = U8NULL;
		gChnlDrvControlConfig[i].CurrentXferTAG = U8NULL;
		
		gChnlDrvControlConfig[i].PM = FALSE;
		if (gChnlConfig[i].ChnlType == CHNL_SATA)
			gChnlDrvControlConfig[i].Support_device = 1;
		else
			gChnlDrvControlConfig[i].Support_device = 2;
		gChnlDrvControlConfig[i].TCQTransferData = FALSE;
	}

	/* initialize TimerID */
	for( i = StartChnl; i <= EndChnl; i++ ){
		ATA_Timeout_TimerID[i]=ZERO_NULL;
		ATA_Port_Multiplier_TimerID[i]=ZERO_NULL;
		BOX_Timeout_TimerID[i]=ZERO_NULL;
		ATA_Polling_Channel_TimerID[i]=ZERO_NULL; 
	}

	/* initialize timeout times*/
	CAMTimeoutTimes[CurAdapter_ID].ATACommand=DEF_ATATIMEOUT;
	CAMTimeoutTimes[CurAdapter_ID].ATAPICommand=DEF_ATAPITIMEOUT;
	CAMTimeoutTimes[CurAdapter_ID].BOX_Check=DEF_CHKREPTIME;	
	CAMTimeoutTimes[CurAdapter_ID].Port_Multiplier_Check=(DEF_CHKREPTIME*DEF_CHKSTATETIME);	


#ifdef _ERR_INJECT_
	
	/* initialize Error Injected SEQ flag */
	ErrInject_SEQ[CurAdapter_ID] = 0;
	ErrInject_SEQ_Timeout[CurAdapter_ID] = 0;
	
	/* initialize Error Injected Task Table */
	for( i = StartChnl; i<= EndChnl; i++ )
	{	
		ErrInj_Buffer_Table[i].head = U16NULL;
		ErrInj_Buffer_Table[i].tail = U16NULL;
		ErrInj_Buffer_Table[i].count = 0;
		ErrInj_Buffer_Table[i].status = FALSE;
	}
#endif

		
	/* enable receive task.*/
	for( i = StartChnl; i <= EndChnl; i++ )
		ATAModuleReceiveTask[i]=TRUE;

#ifdef Encryption
	CAM_EncryptionKeySearch(CurAdapter_ID);
	for( i = StartChnl; i <= EndChnl; i++ )
		TCQ_Encrypt_Flag[i] = FALSE;
#endif

	return(camSUCCESS); /*successful*/
}


void CAM_ControlBMRLongBurst(U8 Adapter_ID, U8 Status)
{
	U32 ldata;

	ReadPortUlong( HostBaseAddr[Adapter_ID], oFFSETHOSTFlashCTL, ldata );

	if(Status == camEnable)
		ldata |= 0x00002000;
	else if(Status == camDisable)
		ldata &= ~0x00002000;

	WritePortUlong(HostBaseAddr[Adapter_ID], oFFSETHOSTFlashCTL, ldata );
	return;
}


void CAM_ResetSerialBusMachine(U8 Adapter_ID)
{
	U16 wdata;
	ReadPortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, wdata);
	wdata&= 0x0F00;
	
	WritePortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, 0x1F0F);
	WritePortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, 0x0F0F);
	
	/* clear Done/Error */
	wdata |= 0x000F;
	WritePortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, wdata);	
	
	return;
}

U8 CAM_Close(PADAPTER_CONFIG_INFO pAdapter)
{
	U8 i;
	U8 CurAdapter_ID;

	if ( pAdapter->Adapter_ID > MAX_ADAPTER_NUM )
		return(camFAIL);
	
	CurAdapter_ID=pAdapter->Adapter_ID;
	
	
	/* For window hibernate, don't disable receive task */
	/* disable receive task.*/
	/*
	for (i=CurAdapter_ID*MAX_ATA_CHNL;i<(CurAdapter_ID+1)*MAX_ATA_CHNL;i++)
		ATAModuleReceiveTask[i]=FALSE;
	*/
	
	/* remove all task */
	/* stop all ata task */
	for (i=CurAdapter_ID*MAX_ATA_CHNL;i<(CurAdapter_ID+1)*MAX_ATA_CHNL;i++){
		stopATAModule(i);
	}

	/* remove all timer */
	for (i=CurAdapter_ID*MAX_ATA_CHNL;i<(CurAdapter_ID+1)*MAX_ATA_CHNL;i++){
		if (ATA_Timeout_TimerID[i] != ZERO_NULL){
			camTimer(CurAdapter_ID,0,(void (*)(void))ATA_Timeout_TimerID[i]);	
			ATA_Timeout_TimerID[i] = ZERO_NULL;
		}
	}

	for (i=CurAdapter_ID*MAX_ATA_CHNL;i<(CurAdapter_ID+1)*MAX_ATA_CHNL;i++){
		if (ATA_Port_Multiplier_TimerID[i] != ZERO_NULL){
			camTimer(CurAdapter_ID,0,(void (*)(void))ATA_Port_Multiplier_TimerID[i]);	
			ATA_Port_Multiplier_TimerID[i] = ZERO_NULL;
		}
	}
	
	for (i=CurAdapter_ID*MAX_ATA_CHNL;i<(CurAdapter_ID+1)*MAX_ATA_CHNL;i++){
		if (BOX_Timeout_TimerID[i] != ZERO_NULL){
			camTimer(CurAdapter_ID,0,(void (*)(void))BOX_Timeout_TimerID[i]);	
			BOX_Timeout_TimerID[i] = ZERO_NULL;
		}
	}

	for (i=CurAdapter_ID*MAX_ATA_CHNL;i<(CurAdapter_ID+1)*MAX_ATA_CHNL;i++){
		if (ATA_Polling_Channel_TimerID[i] != ZERO_NULL){
			camTimer(CurAdapter_ID,0,(void (*)(void))ATA_Polling_Channel_TimerID[i]);	
			ATA_Polling_Channel_TimerID[i] = ZERO_NULL;
		}
	}


	for (i=CurAdapter_ID*MAX_ATA_CHNL;i<(CurAdapter_ID+1)*MAX_ATA_CHNL;i++){
		if (gChnlDrvControlConfig[i].TagCmdFlag & TAG_SEQALLOC) {
			CAMFreeSEQID(CurAdapter_ID,gChnlDrvControlConfig[i].SEQID);
		}
		gChnlDrvControlConfig[i].TagCmdFlag &= ~TAG_SEQALLOC;
		
		gChnlDrvControlConfig[i].SEQID = U8NULL;
	}
			
	return(camSUCCESS); /*successful*/ 
}

void CAM_GetInfo(PCAM_INFO pCAMInfo)
{
	pCAMInfo->VerMajor = CAM_VerMajor;
	pCAMInfo->VerMinor = CAM_VerMinor;
	pCAMInfo->VerOEM   = CAM_VerOEM;
	pCAMInfo->VerBuild = CAM_VerBuild;
	pCAMInfo->CAMFlag  = ACCESS_TYPE;
	pCAMInfo->DeviceID = CAM_DEVICEID;
	pCAMInfo->MinMemSize = MIN_CAM_MEMSIZE;
	pCAMInfo->MaxMemSize = MAX_CAM_MEMSIZE;
	pCAMInfo->PacketSize = CAM_PACKET_SIZE;
	pCAMInfo->NumofATA = MAX_ATA_CHNL;
	pCAMInfo->NumofXOR = MAX_XOR_MODULE;
	pCAMInfo->NumofHDMA = MAX_HDMA_MODULE;
	pCAMInfo->VirMemBase = bit(3)+ bit(4);	/* ask channel driver convert BA#3 for memory-mapped I/O */
						/* ask channel driver convert BA#4 for flash memory access */
	return;	
}

U32 CAM_GetExtraData(U8 Adapter_ID, U8 bID)
{
	U32	lData;
		
	if (!HostBaseAddr[Adapter_ID])
		return(0);
	ReadPortUlong( HostBaseAddr[Adapter_ID],(oFFSETHOSTEXTRA0+(bID << 2)) ,lData);
	
	return(lData);
}

void CAM_SetExtraData(U8 Adapter_ID, U8 bID, U32 value)
{
	if (!HostBaseAddr[Adapter_ID])
		return;
	if ( bID == 0 ) {
		U32	lData;	
		ReadPortUlong( HostBaseAddr[Adapter_ID],(oFFSETHOSTEXTRA0+(bID << 2)) ,lData);
		value &= 0x7FFFFFFF;
		value |= (lData & 0x80000000) ? 0x80000000: 0;
	}
	WritePortUlong( HostBaseAddr[Adapter_ID],(oFFSETHOSTEXTRA0+(bID << 2)) ,value);
	return;
}

void CAM_ATAGetPacketCount(U8 Adapter_ID,PATA_MODULE_PACKET_COUNT pCAM_ATA_Module)
{
	U8 i;
	
	for (i=0;i<MAX_ATA_CHNL;i++)
		pCAM_ATA_Module->ATAModule[i]=ATAPacketQueueCount[Adapter_ID].ATAModule[CAM_ATA_Chnl_OutMAP(i)];
		
	return;
}

void CAM_ATASetPacketCount(U8 Adapter_ID,PATA_MODULE_PACKET_COUNT pCAM_ATA_Module)
{
	U8 i;
	U8 bChannelNumber;

	for (i=0;i<MAX_ATA_CHNL;i++){
		bChannelNumber=CAM_ATA_Chnl_InMAP(i);
		/* can't be zero or large than ATAPacketCount */
		if (pCAM_ATA_Module->ATAModule[i] < 1)
			ATAPacketQueueCount[Adapter_ID].ATAModule[bChannelNumber]=1;
		else if (pCAM_ATA_Module->ATAModule[i] > ATAPacketCount[Adapter_ID])
			ATAPacketQueueCount[Adapter_ID].ATAModule[bChannelNumber]= ATAPacketCount[Adapter_ID];
		else	
			ATAPacketQueueCount[Adapter_ID].ATAModule[bChannelNumber]=pCAM_ATA_Module->ATAModule[i];
	}

	return;
}

void CAM_GetTimeoutTime(U8 Adapter_ID,PMODULE_TIMEOUT pCAM_Timeout)
{

	if ( Adapter_ID > (MAX_ADAPTER_NUM-1) )
		return;
		
	pCAM_Timeout->ATACommand=CAMTimeoutTimes[Adapter_ID].ATACommand;
	pCAM_Timeout->ATAPICommand=CAMTimeoutTimes[Adapter_ID].ATAPICommand;
	pCAM_Timeout->BOX_Check=CAMTimeoutTimes[Adapter_ID].BOX_Check;
	pCAM_Timeout->Port_Multiplier_Check=CAMTimeoutTimes[Adapter_ID].Port_Multiplier_Check;

	return;
}

void CAM_SetTimeoutTime(U8 Adapter_ID,PMODULE_TIMEOUT pCAM_Timeout)
{
	
	if ( Adapter_ID > (MAX_ADAPTER_NUM-1) )
		return;
		
	if (pCAM_Timeout->ATACommand != 0)
		CAMTimeoutTimes[Adapter_ID].ATACommand=pCAM_Timeout->ATACommand;
	if (pCAM_Timeout->ATAPICommand != 0)
		CAMTimeoutTimes[Adapter_ID].ATAPICommand=pCAM_Timeout->ATAPICommand;
	if (pCAM_Timeout->BOX_Check != 0)
		CAMTimeoutTimes[Adapter_ID].BOX_Check=pCAM_Timeout->BOX_Check;
	if (pCAM_Timeout->Port_Multiplier_Check != 0)
		CAMTimeoutTimes[Adapter_ID].Port_Multiplier_Check=pCAM_Timeout->Port_Multiplier_Check;
		
	return;
}

void CAMInitI2CQ(U8 Adapter_ID)
{
	if (Adapter_ID > (MAX_ADAPTER_NUM -1))
		return;
        
	I2CQueueHead[Adapter_ID] = 0;
	I2CQueueTail[Adapter_ID] = 0;
	I2CBusy[Adapter_ID]=FALSE;
		        
	return;
}

U8 PutToI2CQueue(U8 Adapter_ID,PCAM_I2C_QUEUE pI2C_QUEUE)
{
	if (Adapter_ID > (MAX_ADAPTER_NUM -1))
		return (camFAIL);

	if (I2CQueueHead[Adapter_ID] == ((I2CQueueTail[Adapter_ID]+1)%(MAX_I2C_PAC+1)) )
		return (camFAIL); /* over queue OAC */


	if(pI2C_QUEUE->Data || pI2C_QUEUE->Data_CNT){ /* put in tail */
		CAM_I2CQueue[Adapter_ID][I2CQueueTail[Adapter_ID]].callback=pI2C_QUEUE->callback;
		CAM_I2CQueue[Adapter_ID][I2CQueueTail[Adapter_ID]].Data=pI2C_QUEUE->Data;
		CAM_I2CQueue[Adapter_ID][I2CQueueTail[Adapter_ID]].Data_CNT=pI2C_QUEUE->Data_CNT;
		CAM_I2CQueue[Adapter_ID][I2CQueueTail[Adapter_ID]].b_ID=pI2C_QUEUE->b_ID;
		CAM_I2CQueue[Adapter_ID][I2CQueueTail[Adapter_ID]].LED_Type=pI2C_QUEUE->LED_Type;	
		I2CQueueTail[Adapter_ID]++;
		if (I2CQueueTail[Adapter_ID]== (MAX_I2C_PAC+1) )
			I2CQueueTail[Adapter_ID]=0;
	}	
	else{ /* put in head, for timer monitor status */
		if(I2CQueueHead[Adapter_ID] == 0)
			I2CQueueHead[Adapter_ID] = (MAX_I2C_PAC+1);
		I2CQueueHead[Adapter_ID]--;
		CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].callback=pI2C_QUEUE->callback;
		CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].Data=pI2C_QUEUE->Data;
		CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].Data_CNT=pI2C_QUEUE->Data_CNT;
		CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].b_ID=pI2C_QUEUE->b_ID;
		CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].LED_Type=pI2C_QUEUE->LED_Type;	
	}
	return camSUCCESS;
}

U8 GetFromI2CQueue(U8 Adapter_ID)
{
	U8 QueueNumber=0;
	
	if (Adapter_ID > (MAX_ADAPTER_NUM -1))
		return U8NULL;
        		
	if(I2CQueueHead[Adapter_ID]==I2CQueueTail[Adapter_ID])	/* no queue packet */
		return U8NULL;
	
	QueueNumber=I2CQueueHead[Adapter_ID];

	I2CQueueCurrent[Adapter_ID].callback = CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].callback;
	I2CQueueCurrent[Adapter_ID].Data = CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].Data;
	I2CQueueCurrent[Adapter_ID].Data_CNT = CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].Data_CNT;
	I2CQueueCurrent[Adapter_ID].b_ID = CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].b_ID;
	I2CQueueCurrent[Adapter_ID].LED_Type = CAM_I2CQueue[Adapter_ID][I2CQueueHead[Adapter_ID]].LED_Type;	

	I2CQueueHead[Adapter_ID]++;
	
	if (I2CQueueHead[Adapter_ID]== (MAX_I2C_PAC+1) )
		I2CQueueHead[Adapter_ID]=0;	

	return QueueNumber;
}

U8 CAM_StartI2C(U8 Adapter_ID)
{
	U8 QueueNumber=0;
	U16	I2cDelay;
	U8	bData;
	
	if (Adapter_ID > (MAX_ADAPTER_NUM -1))
		return camFAIL;
        		
	if (I2CBusy[Adapter_ID] ==TRUE)
		return camBUSY;
	
	QueueNumber = GetFromI2CQueue(Adapter_ID);
	
	if (QueueNumber==U8NULL)
	{
		/* unmask ERR & Finish and clear ERR & Finish */
		WritePortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, 0x000F);
		return camFREE;
	}
	
	I2CBusy[Adapter_ID] =TRUE;
	
	if(CAM_I2CQueue[Adapter_ID][QueueNumber].Data_CNT || CAM_I2CQueue[Adapter_ID][QueueNumber].Data){
		
		U8 i;
		PU8 Buffer;
	
		/* unmask ERR & Finish and clear ERR & Finish */
		WritePortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, 0x000F);	
		
		/* prepare SBUS buffer */
		Buffer=(PU8)(SerialBusAddr[Adapter_ID].MemVirAddr[0]);
		i=0;
		*(Buffer+i)=((U8)(CAM_I2CQueue[Adapter_ID][QueueNumber].Data) & 0xFE); /* bit 0 : R/W 1/0 */ /* Address */
		i++;
		*(Buffer+i)=0x80; /* Control */
		i++;
		*(Buffer+i)=CAM_I2CQueue[Adapter_ID][QueueNumber].Data_CNT-1; /* Byte Count */
		i++;
		i++;	
		
		*((PU32)(Buffer+i)) = (CAM_I2CQueue[Adapter_ID][QueueNumber].Data>>8);
			
		/* Wait for Series bus busy */	
		for (I2cDelay=0;I2cDelay<MAXDELAYFORI2C;I2cDelay++){
			ReadPortUchar(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, bData );
			if (!(bData & 0x70))	/* can submit task */
				break;
			CAM_ResetSerialBusMachine(Adapter_ID);
		}
		/*if(I2cDelay >= MAXDELAYFORI2C)
			return camFAIL;*/

		/* send serial bus master command */
		WritePortUlong(HostBaseAddr[Adapter_ID], oFFSETSBusMasterBaseAddr, (U32)SerialBusAddr[Adapter_ID].MemPhyAddress[0]);
	}
	else{ /* SATAbox timer event */
		CAM_I2CQueue[Adapter_ID][QueueNumber].callback(&CAM_I2CQueue[Adapter_ID][QueueNumber]);	
	}
	return camSUCCESS;
}

/* MAX write 4 byte(dword) */
U8 CAM_SerialBusWriteDATA(U8 Adapter_ID,PCAM_I2C_QUEUE pI2C_QUEUE)
{
	U8 retunr_value;
	
	retunr_value=PutToI2CQueue(Adapter_ID,pI2C_QUEUE);
	CAM_StartI2C(Adapter_ID);
	
	return (retunr_value);
}

/* 
These two function must be called of the enclosure is workable
CAM_SerialBusWriteDATAPolling()
CAM_SerialBusReadDATAPolling()
*/
/* MAX write 4 byte(dword) */
void CAM_SerialBusWriteDATAPolling(U8 Adapter_ID,U32 data,U8 w_cnt)
{
	U8 i;
	U8 bData=0;
	PU8 Buffer;
	U16 j;
	U8 I2cRetryTime;

	/* mask ERR & Finish and clear ERR & Finish */
	WritePortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, 0x0F0F);	
	
	/* prepare SBUS buffer */
	Buffer=(PU8)(SerialBusAddr[Adapter_ID].MemPollVirAddr[0]);
	i=0;
	*(Buffer+i)=((U8)(data) & 0xFE); /* bit 0 : R/W 1/0 */ /* Address */
	i++;
	*(Buffer+i)=0x80; /* Control */
	i++;
	*(Buffer+i)=w_cnt-1; /* Byte Count */
	i++;
	i++;	
	
	*((PU32)(Buffer+i)) = (U32)(data >> 8);
		

	for(I2cRetryTime=0; I2cRetryTime<5; I2cRetryTime++){

		/* Wait for Series bus busy */	
		for (j=0;j<MAXDELAYFORI2C;j++){
			ReadPortUchar(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, bData );
			if (!(bData & 0x70))	/* can submit task */
				break;
			CAM_ResetSerialBusMachine(Adapter_ID);

		}
		/*if(I2cDelay >= MAXDELAYFORI2C)
			break;*/

		/* send serial bus master command */
		WritePortUlong(HostBaseAddr[Adapter_ID], oFFSETSBusMasterBaseAddr, (U32)SerialBusAddr[Adapter_ID].MemPollPhyAddress[0]);
		
		/* wait return */
		for (j=0;j<MAXDELAYFORI2C;j++){
			ReadPortUchar(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, bData );
			if (bData & 0x0F){	/* FINISH or ERR */
				CAM_ResetSerialBusMachine(Adapter_ID);
				break;
			}
		}
		if(bData & 0x01)	/* Finish */				
			break;
	}		
	return;
}

/* MAX read 4 byte(dword) */
U32 CAM_SerialBusReadDATAPolling(U8 Adapter_ID,U8 index,U8 r_cnt)
{
	U8 bData=0;
	U8 i;
	PU8 Buffer;
	U16	j;
	U8 I2cRetryTime;

	/* mask ERR & Finish and clear ERR & Finish */
	WritePortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, 0x0F0F);
		
	/* prepare SBUS buffer */
	Buffer=(PU8)(SerialBusAddr[Adapter_ID].MemPollVirAddr[0]);
	i=0;
	*(Buffer+i)=(index|0x01); /* bit 0 : R/W 1/0 */ /* Address */
	i++;
	*(Buffer+i)=0x80; /* Control */
	i++;
	*(Buffer+i)=r_cnt; /* Byte Count */
	i++;	
	i++;
	*((PU32)(Buffer+i)) = (U32)SerialBusAddr[Adapter_ID].MemPollPhyAddress[1];
	

	for(I2cRetryTime=0; I2cRetryTime<5; I2cRetryTime++){

		/* Wait for Series bus busy */	
		for (j=0;j<MAXDELAYFORI2C;j++){
			ReadPortUchar(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, bData );
			if (!(bData & 0x70))	/* can submit task */
				break;
			CAM_ResetSerialBusMachine(Adapter_ID);
		}
		/*if(I2cDelay >= MAXDELAYFORI2C)
			break;*/
			
		/* send serial bus master command */
		WritePortUlong(HostBaseAddr[Adapter_ID], oFFSETSBusMasterBaseAddr, (U32)SerialBusAddr[Adapter_ID].MemPollPhyAddress[0]);
		
		/* wait return */
		for (j=0;j<MAXDELAYFORI2C;j++){
			ReadPortUchar(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, bData );
			if (bData & 0x0F){	/* FINISH or ERR */
				CAM_ResetSerialBusMachine(Adapter_ID);
				break;
			}
		}
		if(bData & 0x01)	/* Finish */				
			break;
	}		


	return (*((PU32)(SerialBusAddr[Adapter_ID].MemPollVirAddr[1])));
}

#ifdef Encryption
U8 CAM_EnumEncryptionKey(U8 Adapter_ID)
{
	if (Adapter_ID > (MAX_ADAPTER_NUM-1))
       	return U8NULL;
	
	if(gAdapterConfig[Adapter_ID].Features & WITH_EncryKEY)	
		return camSUCCESS;
	else
		return camFAIL;	
}
#endif

